/* slurp-resource.c
 *
 * Copyright © 2017 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define G_LOG_DOMAIN "slurp-resource"

#include "slurp-resource.h"

/**
 * SECTION:slurp-resource
 * @title: SlurpResource
 * @short_description: Asynchronous resource download using the libsoup package
 * @include: slurp-download.h
 * @see_also: #SoupSession, #SoupRequest, #GMainLoop, #GMainContext,
 *            #GCallback, #GCancellable, #GFile, #GInputStream
 *
 * Downloads files in parallel by running asynchronous file download with the
 * help of libsoup. Each download runs it's own #GMainLoop but refs the current
 * #GMainContext, if one exists using g_main_context_ref_thread_default(). If no
 * current #GMainContext exists then a new one is created. If you plan on using
 * downloads inside of an application you should run that application first
 * either with a #GMainLoop or using #GApplication so I can hook onto the
 * main context.
 **/
/*
static gchar *
supported_schemes = [
    "http",
    "https",
];
*/
struct _SlurpResource
{
    GObject parent_instance;
};

typedef struct
{
    gchar *default_directory;
    gboolean default_overwrite;
    gchar *user_agent;

    GThreadPool *download_resource_thread_pool;
    SoupSession *session;
} SlurpResourcePrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpResource, slurp_resource, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_DEFAULT_DIRECTORY,
    PROP_DEFAULT_OVERWRITE,
    PROP_USER_AGENT,
    N_PROPS
};

static GParamSpec *properties[N_PROPS];

enum {
    ADDED,
    REMOVED,
    FINISHED,
    CANCELLED,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

/**
 * download_resource_create_path:
 * @self: a #SlurpResource
 * @uri: a string to a uri resource
 * @path: a string to a valid path
 * @cancellable: a #GCancellable
 *
 * Create a file path to save the resource from @uri. The finial path is created
 * depending on the values of @uri, @path, and the
 * #SlurpResource::default-directory.
 *
 * Returns: a %const string file path
 */
static const gchar *
download_resource_create_path (SlurpResource *self,
                               const gchar   *uri,
                               const gchar   *path,
                               GCancellable  *cancellable)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);
    gchar *file_path = NULL;
    gchar *file_name = NULL;
    gchar *file = NULL;

    if (uri == NULL)
        file_name = g_uuid_string_random ();
    else
        file_name = g_path_get_basename (uri);

    if (path == NULL)
        file_path = g_strdup (priv->default_directory);
    else
        file_path = g_strdup (path);

    if (g_str_has_suffix (file_path, "/"))
    {
        if (!g_file_test (file_path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))
        {
            g_mkdir_with_parents (file_path, 0700);
        }
    }

    file = g_build_filename (file_path, file_name, NULL);

    g_free (file_path);
    g_free (file_name);

    return file;
}

/**
 * download_resource_can_write_path:
 * @self: a #SlurpResource
 * @path: a string to a valid path
 * @overwrite: %TRUE or %FALSE if file can be overwritten
 * @cancellable: a #GCancellable
 *
 * Check if @path is writable.
 *
 * Returns: %TRUE if path is writable.
 */
static gboolean
download_resource_can_write_path (SlurpResource *self,
                                  const gchar   *path,
                                  gboolean       overwrite,
                                  GCancellable  *cancellable)
{
    GFile *file = g_file_new_for_path (path);
    gboolean retval;

    retval = (!overwrite && g_file_query_exists (file, cancellable));

    g_object_unref (file);

    return retval;
}














typedef struct _SlurpResourceTaskData {
    gchar *uuid;
    const gchar *uri;
    const gchar *path;
    gboolean overwrite;

    GFile *file;
    gchar buffer[16384]; // 16 * 1024
    GOutputStream *output;

    guint64 total_bytes;
    guint64 downloaded_bytes;
    guint64 last_progress_time;
} SlurpResourceTaskData;

static void
slurp_resource_task_data_free (SlurpResourceTaskData *data)
{
    g_free (data->uuid);
    g_clear_pointer (&data->file, g_object_unref);
    g_free (data->buffer);
    g_clear_pointer (&data->output, g_object_unref);

    g_slice_free (SlurpResourceTaskData, data);
}

static void
_download_resource_from_uri_async ()
{
}

gboolean
slurp_resource_download (SlurpResource  *self,
                         const gchar    *uri,
                         const gchar    *path,
                         gboolean        overwrite,
                         GCancellable   *cancellable,
                         GError        **error)
{
    g_print ("Downloading...\n");

    // Check URI
    // Check path
    // Check write
    // Start download

    soup_request_send_async (SOUP_REQUEST(request),
                             cancellable,
                             download_resource_from_uri_async_cb,
                             resource);

    return TRUE;
}

static void
slurp_resource_download_in_thread_func (gpointer data,      /* task_data from g_task_new() */
                                        gpointer user_data) /* user_data from g_thread_pool_new() */
{
    g_return_if_fail (G_IS_TASK (data) && data != NULL);
    g_return_if_fail (SLURP_IS_RESOURCE (user_data));

    GTask *task = G_TASK (data);
    SlurpResource *self = SLURP_RESOURCE (user_data);
    SlurpResourceTaskData *task_data = g_task_get_task_data (task);
    GCancellable *cancellable = g_task_get_cancellable (task);
    GError *error = NULL;
    gboolean retval;

    retval = slurp_resource_download (self,
                                      task_data->uri,
                                      task_data->path,
                                      task_data->overwrite,
                                      cancellable,
                                      &error);

    if (!retval)
    {
        /*
         * If the user cancelled the download then clean-up the file
         */
        if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
            g_file_delete (task_data->file, NULL, NULL);

        g_task_return_error (task, error);
        return;
    }

    g_task_return_boolean (task, TRUE);
}

void
slurp_resource_download_async_full (SlurpResource       *self,
                                    const gchar         *uri,
                                    const gchar         *file,
                                    gboolean             overwrite
                                    GCancellable        *cancellable,
                                    GCallback            p_handler,
                                    gpointer             p_user_data,
                                    GAsyncReadyCallback  callback,
                                    gpointer             user_data)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);
    SlurpResourceTaskData *task_data = NULL;
    GTask *task = NULL;
    GError *error = NULL;
    gboolean retval = FALSE;

    task_data = g_slice_new0 (SlurpResourceTaskData);
    task_data->uuid = g_uuid_string_random ();
    task_data->uri = g_strdup (uri);
    task_data->overwrite = overwrite;
    task_data->path = g_strdup (path);
    task_data->file = NULL;

    task_data->total_bytes = 0;
    task_data->downloaded_bytes = 0;
    task_data->last_progress_time = g_get_monotonic_time ();

    g_debug ("Resource added to download queue with UUID \"%s\"\n from URI \"%s\"",
             task_data->uuid,
             task_data->uri);

    task = g_task_new (self, cancellable, callback, user_data);
    g_task_set_source_tag (task, slurp_resource_resource_from_uri_async);
    g_task_set_task_data (task, task_data, (GDestroyNotify) slurp_resource_task_data_free);

    retval = g_thread_pool_push (priv->download_resource_thread_pool,
                                 g_object_ref (task),
                                 &error);

    if (!retval)
    {
        g_debug ("Unable to add resource with UUID \"%s\" to download queue", task_data->uuid);
        g_object_unref (task);
        return;
    }
}

gboolean
slurp_resource_download_finish (SlurpResource  *self,
                                GAsyncResult   *result,
                                GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (G_IS_TASK (result), FALSE);
    g_return_val_if_fail (g_task_is_valid (result, self), FALSE);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), FALSE);
    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

    GTask *task = G_TASK (result);

    g_return_val_if_fail (g_task_get_source_tag (task) == slurp_resource_resource_from_uri_async, FALSE);

    return g_task_propagate_boolean (task, error);
}

/*
static void
slurp_resource_resource_from_uri_cb (GObject      *object,
                                     GAsyncResult *result,
                                     gpointer      user_data)
{
    g_return_if_fail (SLURP_IS_RESOURCE (object));
    g_return_if_fail (G_IS_ASYNC_RESULT (result));

    SlurpResource *self = SLURP_RESOURCE (object);
    gboolean retval;
    GError *error = NULL;

    retval = slurp_resource_resource_from_uri_finish (self, result, &error);

    if (error != NULL)
    {
        if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
            goto out;

        // TODO: make better
        g_warning ("Unable to download resource: %s", error->message);
    }

    if (!retval)
    {
        //TODO: Trigger failed download signal
        goto out;
    }

    //TODO: Trigger success download signal

out:
    return;
}
*/

static void
slurp_resource_finalize (GObject *object)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    g_clear_pointer (&priv->default_directory, g_free);
    g_clear_pointer (&priv->user_agent, g_free);

    G_OBJECT_CLASS (slurp_resource_parent_class)->finalize (object);
}

static void
slurp_resource_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_DEFAULT_DIRECTORY:
            g_value_set_string (value, priv->default_directory);
            break;

        case PROP_DEFAULT_OVERWRITE:
            g_value_set_boolean (value, priv->default_overwrite);
            break;

        case PROP_USER_AGENT:
            g_value_set_string (value, priv->user_agent);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_resource_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_DEFAULT_DIRECTORY:
            priv->default_directory = g_value_dup_string (value);
            break;

        case PROP_DEFAULT_OVERWRITE:
            priv->default_overwrite = g_value_get_boolean (value);
            break;

        case PROP_USER_AGENT:
            priv->user_agent = g_value_dup_string (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_resource_class_init (SlurpResourceClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_resource_finalize;
    object_class->get_property = slurp_resource_get_property;
    object_class->set_property = slurp_resource_set_property;

	/**
	 * SlurpResource:directory:
	 *
	 * The default directory to save resources into.
     *
     * Default: $XDG_USER_DIR/RESOURCE (ex. $HOME/Downloads)
	 */
    properties[PROP_DEFAULT_DIRECTORY] =
        g_param_spec_string ("directory",
                             "Default directory",
                             "The default directory to save resources into.",
                             g_get_user_special_dir (G_USER_DIRECTORY_RESOURCE),
                             G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * SlurpResource:overwrite:
	 *
	 * Should files be overwritten by default.
     *
     * Recommended: %FALSE
	 */
    properties[PROP_DEFAULT_OVERWRITE] =
        g_param_spec_boolean ("overwrite",
                              "Overwrite files",
                              "Should files be overwritten by default? Default: FALSE",
                              FALSE,
                              G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * SlurpResource:user-agent:
	 *
	 * The user agent to pass when establishing a connection to the resource.
     *
     * Default: libslurp
	 */
    properties[PROP_USER_AGENT] =
        g_param_spec_string ("user-agent",
                              "User Agent",
                              "The user agent to pass when establishing a connection to the resource.",
                              "libslurp",
                              G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY);

    g_object_class_install_properties (object_class,
                                       N_PROPS,
                                       properties);

    signals[ADDED] =
        g_signal_new ("added",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);

    signals[REMOVED] =
        g_signal_new ("removed",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);

    signals[FINISHED] =
        g_signal_new ("finished",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);

    signals[CANCELLED] =
        g_signal_new ("cancelled",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      G_TYPE_STRING);
}

static void
slurp_resource_init (SlurpResource *self)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    priv->download_resource_thread_pool =
        g_thread_pool_new (slurp_resource_download_in_thread_func,
                           self,
                           1,
                           FALSE,
                           NULL);

    priv->session =
        soup_session_new_with_options (SOUP_SESSION_USER_AGENT, priv->user_agent,
                                       SOUP_SESSION_SSL_USE_SYSTEM_CA_FILE, TRUE,
                                       SOUP_SESSION_USE_THREAD_CONTEXT, TRUE,
                                       SOUP_SESSION_TIMEOUT, 60,
                                       SOUP_SESSION_IDLE_TIMEOUT, 60,
                                       NULL);
}

SlurpResource *
slurp_resource_new (void)
{
    return g_object_new (SLURP_TYPE_RESOURCE, NULL);
}
