/* slurp-download.c
 *
 * Copyright © 2017 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define G_LOG_DOMAIN "slurp-download"

#include "slurp-download.h"

/**
 * SECTION:slurp-download
 * @title: SlurpDownload
 * @short_description: Asynchronous resource downloader using the libsoup library
 * @include: slurp-download.h
 * @see_also: #SoupSession, #SoupRequest, #GCallback, #GCancellable, #GFile,
 *            #GInputStream
 *
 * Asynchronous resource downloader using the libsoup library which offers
 * addded features over using plain `soup_*` calls.
 **/

typedef struct
{
    gchar *default_directory;
    gboolean default_overwrite;
    gchar *user_agent;

    SoupSession *session;
} SlurpDownloadPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpDownload, slurp_download, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_SESSION,
    PROP_USER_AGENT,
    PROP_DEFAULT_DIRECTORY,
    PROP_DEFAULT_OVERWRITE,
    NUM_PROPERTIES
};

static GParamSpec *properties[NUM_PROPERTIES];

GQuark
slurp_download_error_quark (void)
{
	static GQuark quark = 0;

	if (G_UNLIKELY (quark == 0))
	{
		quark = g_quark_from_static_string ("slurp-download-error");
	}

	return quark;
}

void
slurp_download_resource_splice_streams_close_async_cb (GObject      *object,
                                                       GAsyncResult *result,
                                                       gpointer      user_data)
{
    g_autoptr (GInputStream) input = NULL;
    g_autoptr (GTask) task = NULL;
    g_autoptr (GError) error = NULL;
    SlurpDownloadReadData *data = NULL;

    g_return_if_fail (G_IS_INPUT_STREAM (object));
    g_return_if_fail (G_IS_TASK (user_data));

    input = G_INPUT_STREAM (object);
    task = G_TASK (user_data);
    data = g_task_get_task_data (task);

    g_debug ("Download (%s) closing GInputStream",
             slurp_resource_get_source (data->resource));

    if (!g_input_stream_close_finish (input, result, &error))
    {
        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_INPUT_STREAM_CLOSE_FAILED,
                                 "Failed to close GInputStream: %s",
                                 error->message);
    }

    slurp_resource_force_update_progress (data->resource);

    g_task_return_boolean (task, TRUE);
}

void
slurp_download_resource_splice_streams_async_cb (GObject      *object,
                                                 GAsyncResult *result,
                                                 gpointer      user_data)
{
    g_autoptr (GInputStream) input = NULL;
    g_autoptr (GTask) task = NULL;
    g_autoptr (GError) error = NULL;
    SlurpDownloadReadData *data = NULL;
    gsize nread;

    g_return_if_fail (G_IS_INPUT_STREAM (object));
    g_return_if_fail (G_IS_TASK (user_data));

    input = G_INPUT_STREAM (object);
    task = G_TASK (user_data);
    data = g_task_get_task_data (task);

    nread = g_input_stream_read_finish (G_INPUT_STREAM (object), result, &error);

    g_debug ("Download (%s) reading new chunk from GInputStream",
             slurp_resource_get_source (data->resource));

    slurp_resource_update_progress (data->resource);

    if (nread == -1 || nread == 0 || error)
    {
        g_debug ("Download (%s) end of GInputStream",
                 slurp_resource_get_source (data->resource));

        if (error)
        {
            g_task_return_new_error (task,
                                     SLURP_DOWNLOAD_ERROR,
                                     SLURP_DOWNLOAD_ERROR_NO_OUTPUT_STREAM,
                                     "No GOutputStream object.");
        }

        g_input_stream_close_async (g_object_ref (input),
                                    G_PRIORITY_DEFAULT,
                                    g_task_get_cancellable (task),
                                    slurp_download_resource_splice_streams_close_async_cb,
                                    g_object_ref (task));

        return;
    }

    if (data->output)
    {
        gboolean written;
        gsize n_written;

        written = g_output_stream_write_all (G_OUTPUT_STREAM (data->output),
                                             data->buffer,
                                             nread,
                                             &n_written,
                                             NULL, // TODO: cancellable
                                             &error);

        g_debug ("Download (%s) wrote %s \"%lu\" bytes",
                 slurp_resource_get_source (data->resource),
                 written ? "successfully" : "unsuccessfully",
                 n_written);

        if (error)
        {
            g_task_return_new_error (task,
                                     SLURP_DOWNLOAD_ERROR,
                                     SLURP_DOWNLOAD_ERROR_OUTPUT_STREAM_WRITE_FAILED,
                                     "Stream write failed for \"%s\": %s",
                                     "SOURCE",
                                     error->message);
        }

        slurp_resource_add_downloaded_bytes (data->resource, n_written);

        if (slurp_resource_get_downloaded_bytes (data->resource) > slurp_resource_get_total_bytes (data->resource))
        {
            slurp_resource_set_total_bytes (data->resource,
                                            slurp_resource_get_downloaded_bytes (data->resource));
        }

        if (!written)
        {
            g_input_stream_close_async (g_object_ref (input),
                                        G_PRIORITY_DEFAULT,
                                        g_task_get_cancellable (task),
                                        slurp_download_resource_splice_streams_close_async_cb,
                                        g_object_ref (task));
            return;
        }
    }
    else
    {
        g_debug ("Download (%s) GOutputStream no longer exists, quit",
                 slurp_resource_get_source (data->resource));

        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_NO_OUTPUT_STREAM,
                                 "No GOutputStream object.");

        return;
    }

    g_input_stream_read_async (g_object_ref (input),
                               data->buffer,
                               sizeof (data->buffer),
                               G_PRIORITY_DEFAULT,
                               g_task_get_cancellable (task),
                               slurp_download_resource_splice_streams_async_cb,
                               g_object_ref (task));
}

void
slurp_download_resource_splice_streams_async (SlurpDownload       *self,
                                              GInputStream        *input,
                                              GOutputStream       *output,
                                              GCancellable        *cancellable,
                                              GAsyncReadyCallback  callback,
                                              gpointer             user_data)
{
    g_autoptr (GTask) parent_task = NULL;
    g_autoptr (GTask) task = NULL;
    SlurpResource *resource = NULL;
    SlurpDownloadReadData *data = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (G_IS_INPUT_STREAM (input));
    g_return_if_fail (G_IS_OUTPUT_STREAM (output));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
    g_return_if_fail (G_IS_TASK (user_data));

    parent_task = G_TASK (user_data);
    resource = g_task_get_task_data (parent_task);

    task = g_task_new (self, cancellable, callback, g_object_ref (user_data));
    g_task_set_source_tag (task, slurp_download_resource_splice_streams_async);

    data = slurp_download_read_data_new ();
    data->resource = g_object_ref (resource);
    data->input = g_object_ref (input);
    data->output = g_object_ref (output);
    g_task_set_task_data (task, data, g_object_unref);

    g_input_stream_read_async (g_object_ref (input),
                               data->buffer,
                               sizeof (data->buffer),
                               G_PRIORITY_DEFAULT,
                               g_task_get_cancellable (task),
                               slurp_download_resource_splice_streams_async_cb,
                               g_object_ref (task));
}

gboolean
slurp_download_resource_splice_streams_finish (GObject       *object,
                                               GAsyncResult  *result,
                                               GError       **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (object), FALSE);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), FALSE);
    g_return_val_if_fail (g_task_is_valid (result, object), FALSE);

    return g_task_propagate_boolean (G_TASK (result), error);
}

void
slurp_download_resource_get_input_stream_async_cb (GObject      *object,
                                                   GAsyncResult *result,
                                                   gpointer      user_data)
{
    g_autoptr (SoupRequest) request = NULL;
    g_autoptr (GTask) task = NULL;
    g_autoptr (GInputStream) stream = NULL;
    g_autoptr (GError) error = NULL;
    SlurpResource *resource = NULL;

    g_return_if_fail (SOUP_IS_REQUEST (object));
    g_return_if_fail (G_IS_TASK (user_data));

    request = SOUP_REQUEST (object);
    task = G_TASK (user_data);
    resource = g_task_get_task_data (task);

    stream = soup_request_send_finish (request, result, &error);

    if (!stream)
    {
        g_task_return_error (task, error);
        return;
    }

    slurp_resource_set_total_bytes (g_object_ref (resource),
                                    (gint64) soup_request_get_content_length (g_object_ref (request)));

    g_task_return_pointer (task, g_object_ref (stream), g_object_unref);
}

void
slurp_download_resource_get_input_stream_async (SlurpDownload       *self,
                                                GCancellable        *cancellable,
                                                GAsyncReadyCallback  callback,
                                                gpointer             user_data)
{
    SlurpDownloadPrivate *priv = NULL;
    g_autoptr (GTask) parent_task = NULL;
    g_autoptr (GTask) task = NULL;
    SlurpResource *resource = NULL;
    g_autoptr (SoupSession) session = NULL;
    g_autoptr (SoupRequest) request = NULL;
    g_autoptr (GError) error = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
    g_return_if_fail (G_IS_TASK (user_data));

    priv = slurp_download_get_instance_private (self);
    parent_task = G_TASK (user_data);
    resource = g_task_get_task_data (parent_task);

    task = g_task_new (self, cancellable, callback, g_object_ref (user_data));
    g_task_set_source_tag (task, slurp_download_resource_get_input_stream_async);
    g_task_set_task_data (task, g_object_ref (resource), g_object_unref);

    if (!priv->session)
    {
        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_NO_SOUP_SESSION,
                                 "No SoupSession set. Either failed to create SoupSession in SlurpDownload or NULL session passed.");

        return;
    }

    session = g_object_ref (priv->session);
    request = soup_session_request (session,
                                    slurp_resource_get_source (resource),
                                    &error);

    if (!request)
    {
        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_NO_SOUP_REQUEST,
                                 "Failed to create SoupRequest for source URI \"%s\".",
                                 g_strdup (slurp_resource_get_source (resource)));

        return;
    }

    soup_request_send_async (g_object_ref (request),
                             cancellable,
                             slurp_download_resource_get_input_stream_async_cb,
                             g_object_ref (task));
}

GInputStream *
slurp_download_resource_get_input_stream_finish (GObject       *object,
                                                 GAsyncResult  *result,
                                                 GError       **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (object), NULL);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), NULL);
    g_return_val_if_fail (g_task_is_valid (result, object), NULL);

    return g_task_propagate_pointer (G_TASK (result), error);
}

void
slurp_download_resource_save_from_input_stream_async_cb (GObject      *object,
                                                         GAsyncResult *result,
                                                         gpointer      user_data)
{
    g_autoptr (GTask) task = NULL;
    g_autoptr (GError) error = NULL;
    gboolean retval;

    g_return_if_fail (SLURP_IS_DOWNLOAD (object));
    g_return_if_fail (G_IS_TASK (user_data));

    task = G_TASK (user_data);

    retval = slurp_download_resource_splice_streams_finish (object, result, &error);

    if (!retval)
    {
        g_task_return_error (task, error);
        return;
    }

    g_task_return_boolean (task, TRUE);
}

void
slurp_download_resource_save_from_input_stream_async (GObject      *object,
                                                      GAsyncResult *result,
                                                      gpointer      user_data)
{
    SlurpDownload *self = NULL;
    SlurpResource *resource = NULL;
    GCancellable *cancellable = NULL;
    g_autoptr (GInputStream) input = NULL;
    g_autoptr (GOutputStream) output = NULL;
    g_autoptr (GTask) task = NULL;
    g_autoptr (GError) error = NULL;
    g_autoptr (GFile) file = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (object));
    g_return_if_fail (G_IS_ASYNC_RESULT (result));
    g_return_if_fail (G_IS_TASK (user_data));
    g_return_if_fail (g_task_is_valid (user_data, object));
    g_return_if_fail (SLURP_IS_RESOURCE (g_task_get_task_data (user_data)));

    self = SLURP_DOWNLOAD (object);
    task = G_TASK (user_data);
    resource = g_task_get_task_data (task);
    cancellable = g_task_get_cancellable (task);

    input = slurp_download_resource_get_input_stream_finish (object, result, &error);

    if (!input)
    {
        if (error)
        {
            g_task_return_new_error (task,
                                     SLURP_DOWNLOAD_ERROR,
                                     SLURP_DOWNLOAD_ERROR_NO_INPUT_STREAM,
                                     "Failed to create GInputStream for source URI \"%s\": %s",
                                     g_strdup (slurp_resource_get_source (resource)),
                                     error->message);
        }
        else
        {
            g_task_return_new_error (task,
                                     SLURP_DOWNLOAD_ERROR,
                                     SLURP_DOWNLOAD_ERROR_NO_INPUT_STREAM,
                                     "Failed to create GInputStream for source URI \"%s\"",
                                     g_strdup (slurp_resource_get_source (resource)));
        }

        return;
    }

    // Get target name
    //
    // if (file exists <target> && overwrite)
    //  all good
    // else
    //  create_target (<target>(<version>).ext)
    //  slurp_resource_set_target (resource, target);
    //
    // (<target>.part)
    // if (file exists (<target>.part))
    //  iostream = get seek
    // else
    //  iostream = g_file_new_for_uri (<target>)
    //
    // output = g_io_stream_get_output_stream (iostream)

    //file = g_file_new_tmp ("slurp-download-XXXXXX", &iostream, &error);
    //output = g_io_stream_get_output_stream (iostream);
    //
    // g_file_create_readwrite
    // g_file_replace_readwrite

    file = g_file_new_for_uri (slurp_resource_get_target (resource));

    output = G_OUTPUT_STREAM (g_file_replace (file,
                                              NULL,
                                              FALSE,
                                              //G_FILE_CREATE_NONE,
                                              G_FILE_CREATE_REPLACE_DESTINATION,
                                              g_task_get_cancellable (task),
                                              &error));

    if (error)
        g_print ("ERROR: %s\n", error->message);

    if (!output)
    {
        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_NO_OUTPUT_STREAM,
                                 "Failed to create GOutputStream for target URI \"%s\": %s",
                                 g_strdup (slurp_resource_get_target (resource)),
                                 error->message);
        return;
    }

    // Do GFile checks

    g_debug ("### Streams passed checks");

    /*
     * We don't use g_output_stream_splice_async() because we want to update
     * the progress to the passed SlurpResource.
     */
    slurp_download_resource_splice_streams_async (self,
                                                  g_object_ref (input),
                                                  g_object_ref (output),
                                                  cancellable,
                                                  slurp_download_resource_save_from_input_stream_async_cb,
                                                  g_object_ref (task));

}

gboolean
slurp_download_resource_save_from_input_stream_finish (GObject       *object,
                                                       GAsyncResult  *result,
                                                       GError       **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (object), FALSE);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), FALSE);
    g_return_val_if_fail (g_task_is_valid (result, object), FALSE);

    return g_task_propagate_boolean (G_TASK (result), error);
}

void
slurp_download_resource_async (SlurpDownload       *self,
                               SlurpResource       *resource,
                               GCancellable        *cancellable,
                               GAsyncReadyCallback  callback,
                               gpointer             user_data)
{
    g_autoptr (GTask) task = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (SLURP_IS_RESOURCE (resource));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));

    task = g_task_new (self, cancellable, callback, user_data);
    g_task_set_source_tag (task, slurp_download_resource_async);
    g_task_set_task_data (task, g_object_ref (resource), (GDestroyNotify) g_object_unref);

    slurp_download_resource_get_input_stream_async (self,
                                                    cancellable,
                                                    slurp_download_resource_save_from_input_stream_async,
                                                    g_object_ref (task));
}

gboolean
slurp_download_resource_finish (GObject       *object,
                                GAsyncResult  *result,
                                GError       **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (object), FALSE);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), FALSE);
    g_return_val_if_fail (g_task_is_valid (result, object), FALSE);

    return g_task_propagate_boolean (G_TASK (result), error);
}

static void
slurp_download_finalize (GObject *object)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    g_clear_pointer (&priv->default_directory, g_free);
    g_clear_pointer (&priv->user_agent, g_free);
    g_clear_pointer (&priv->session, g_object_unref);

    G_OBJECT_CLASS (slurp_download_parent_class)->finalize (object);
}

static void
slurp_download_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SESSION:
            g_value_set_object (value, priv->session);
            break;

        case PROP_USER_AGENT:
            g_value_set_string (value, priv->user_agent);
            break;

        case PROP_DEFAULT_DIRECTORY:
            g_value_set_string (value, priv->default_directory);
            break;

        case PROP_DEFAULT_OVERWRITE:
            g_value_set_boolean (value, priv->default_overwrite);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SESSION:
            priv->session = g_value_dup_object (value);
            break;

        case PROP_USER_AGENT:
            priv->user_agent = g_value_dup_string (value);
            break;

        case PROP_DEFAULT_DIRECTORY:
            priv->default_directory = g_value_dup_string (value);
            break;

        case PROP_DEFAULT_OVERWRITE:
            priv->default_overwrite = g_value_get_boolean (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_class_init (SlurpDownloadClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_download_finalize;
    object_class->get_property = slurp_download_get_property;
    object_class->set_property = slurp_download_set_property;

	/**
	 * SlurpDownload:session:
	 *
	 * A SoupSession which Slurp can use to download resources.
     *
     * Note: Make sure to ref your #SoupSession before passing it.
     *
     * g_object_new():
     * |[
     * g_object_new (SLURP_TYPE_DOWNLOAD,
     *               "session", g_object_ref (session),
     *               NULL);
     * ]|
     *
     * slurp_download_new_with_session():
     * |[
     * SlurpDownload *download;
     * download = slurp_download_new_with_session (g_object_ref (session));
     * ]|
	 */
    properties[PROP_SESSION] =
        g_param_spec_object ("session",
                             "A SoupSession",
                             "A SoupSession which Slurp can use to download resources.",
                             SOUP_TYPE_SESSION,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpDownload:user-agent:
	 *
	 * The user agent to pass when establishing a connection to the resource.
     *
     * Default: libslurp
	 */
    // TODO: Add new user-agent if session is passed
    properties[PROP_USER_AGENT] =
        g_param_spec_string ("user-agent",
                              "User Agent",
                              "The user agent to pass when establishing a connection to the resource.",
                              "slurp-download",
                              G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpDownload:directory:
	 *
	 * The default directory to save resources into.
     *
     * Default: $XDG_USER_DIR/DOWNLOAD (ex. $HOME/Downloads)
	 */
    properties[PROP_DEFAULT_DIRECTORY] =
        g_param_spec_string ("directory",
                             "Default directory",
                             "The default directory to save resources into.",
                             g_get_user_special_dir (G_USER_DIRECTORY_DOWNLOAD),
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * SlurpDownload:overwrite:
	 *
	 * Should files be overwritten by default.
     *
     * Recommended: %FALSE
	 */
    properties[PROP_DEFAULT_OVERWRITE] =
        g_param_spec_boolean ("overwrite",
                              "Overwrite files",
                              "Should files be overwritten by default? Default: FALSE",
                              FALSE,
                              G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);

    g_object_class_install_properties (object_class,
                                       NUM_PROPERTIES,
                                       properties);

}

static void
slurp_download_init (SlurpDownload *self)
{
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    if (!priv->session)
    {
        priv->session =
            soup_session_new_with_options (SOUP_SESSION_USER_AGENT, priv->user_agent,
                                           SOUP_SESSION_SSL_USE_SYSTEM_CA_FILE, TRUE,
                                           SOUP_SESSION_USE_THREAD_CONTEXT, TRUE,
                                           SOUP_SESSION_TIMEOUT, 60,
                                           SOUP_SESSION_IDLE_TIMEOUT, 60,
                                           NULL);
    }
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
SlurpDownload *
slurp_download_new (void)
{
    return g_object_new (SLURP_TYPE_DOWNLOAD, NULL);
}

SlurpDownload *
slurp_download_new_with_session (SoupSession *session)
{
    return g_object_new (SLURP_TYPE_DOWNLOAD,
                         "session", session,
                         NULL);
}