/* slurp-resource.c
 *
 * Copyright © 2017 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define G_LOG_DOMAIN "slurp-resource"

#include "slurp-resource.h"

typedef struct {
    gchar *source_directory;
    gchar *source_filename;
    gchar *target_directory;
    gchar *target_filename;
    gboolean overwrite;
    gint64 total_bytes;
    gint64 downloaded_bytes;
    gint64 last_progress_time;
    gint64 last_progress_time_downloaded_bytes;
    gint64 download_speed;
    gint64 estimated_time;
} SlurpResourcePrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpResource, slurp_resource, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_SOURCE,
    PROP_TARGET,
    PROP_OVERWRITE,
    PROP_TOTAL_BYTES,
    PROP_DOWNLOADED_BYTES,
    PROP_LAST_PROGRESS_TIME,
    PROP_DOWNLOAD_SPEED,
    PROP_ESTIMATED_TIME,
    NUM_PROPERTIES
};

static GParamSpec *properties[NUM_PROPERTIES];

enum {
    PROGRESS,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

static gchar *
_resource_get_source (SlurpResource *self)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), NULL);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    if (priv->source_directory == NULL && priv->source_filename == NULL)
        return NULL;

    return g_build_filename (priv->source_directory, priv->source_filename, NULL);
}

static void
_resource_set_source (SlurpResource *self,
                      const gchar   *uri)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    if (slurp_utils_uri_is_directory (uri))
    {
        priv->source_directory = g_strdup (uri);
        priv->source_filename = NULL;

        return;
    }

    priv->source_directory = g_path_get_dirname (uri);
    priv->source_filename = g_path_get_basename (uri);
}

static gchar *
_resource_get_target (SlurpResource *self)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), NULL);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    return g_build_filename (priv->target_directory, priv->target_filename, NULL);
}

static void
_resource_set_target (SlurpResource *self,
                      const gchar   *uri)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    if (slurp_utils_uri_is_directory (uri))
    {
        priv->target_directory = g_strdup (uri);

        if (priv->source_filename != NULL)
            priv->target_filename = g_strdup (priv->source_filename);
        else
            priv->target_filename = NULL;

        return;
    }

    priv->target_directory = g_path_get_dirname (uri);
    priv->target_filename = g_path_get_basename (uri);
}

static void
_resource_calculate_progress (SlurpResource *self,
                              gboolean       force)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    if (priv->last_progress_time == -1)
    {
        g_debug ("Last progress time not set (-1), can't update progress.");
        priv->last_progress_time = g_get_monotonic_time ();
        priv->last_progress_time_downloaded_bytes = priv->downloaded_bytes;
        return;
    }

    if (force || g_get_monotonic_time () - priv->last_progress_time > 1 * G_USEC_PER_SEC)
    {
        if (priv->downloaded_bytes == -1 || priv->last_progress_time_downloaded_bytes == -1)
        {
            priv->download_speed = -1;
        }
        else
        {
            priv->download_speed =
                (priv->downloaded_bytes - priv->last_progress_time_downloaded_bytes);
        }

        g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_DOWNLOAD_SPEED]);

        if (priv->total_bytes > 0)
        {
            if (priv->download_speed > 0)
            {
                 priv->estimated_time =
                    (priv->total_bytes / priv->download_speed);
            }
            else
            {
                priv->estimated_time = -1;
            }

            g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_ESTIMATED_TIME]);
        }

        priv->last_progress_time = g_get_monotonic_time ();
        priv->last_progress_time_downloaded_bytes = priv->downloaded_bytes;
        g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_LAST_PROGRESS_TIME]);

        g_signal_emit (G_OBJECT (self), signals[PROGRESS], 0,
                       priv->downloaded_bytes,
                       priv->total_bytes,
                       priv->download_speed,
                       priv->estimated_time,
                       NULL);
    }
}

static void
slurp_resource_finalize (GObject *object)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    g_clear_pointer (&priv->source_directory, g_free);
    g_clear_pointer (&priv->source_filename, g_free);
    g_clear_pointer (&priv->target_directory, g_free);
    g_clear_pointer (&priv->target_filename, g_free);

    G_OBJECT_CLASS (slurp_resource_parent_class)->finalize (object);
}

static void
slurp_resource_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SOURCE:
            g_value_set_string (value, _resource_get_source (self));
            break;

        case PROP_TARGET:
            g_value_set_string (value, _resource_get_target (self));
            break;

        case PROP_OVERWRITE:
            g_value_set_boolean (value, priv->overwrite);
            break;

        case PROP_TOTAL_BYTES:
            g_value_set_uint64 (value, priv->total_bytes);
            break;

        case PROP_DOWNLOADED_BYTES:
            g_value_set_uint64 (value, priv->downloaded_bytes);
            break;

        case PROP_LAST_PROGRESS_TIME:
            g_value_set_uint64 (value, priv->last_progress_time);
            break;

        case PROP_DOWNLOAD_SPEED:
            g_value_set_uint64 (value, priv->download_speed);
            break;

        case PROP_ESTIMATED_TIME:
            g_value_set_uint64 (value, priv->estimated_time);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_resource_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SOURCE:
            _resource_set_source (self, g_value_dup_string (value));
            break;

        case PROP_TARGET:
            _resource_set_target (self, g_value_dup_string (value));
            break;

        case PROP_OVERWRITE:
            priv->overwrite = g_value_get_boolean (value);
            break;

        case PROP_TOTAL_BYTES:
            priv->total_bytes = g_value_get_uint64 (value);
            break;

        case PROP_DOWNLOADED_BYTES:
            priv->downloaded_bytes = g_value_get_uint64 (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_resource_class_init (SlurpResourceClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_resource_finalize;
    object_class->get_property = slurp_resource_get_property;
    object_class->set_property = slurp_resource_set_property;

	/**
	 * SlurpResource:source:
	 *
	 *
	 */
    properties[PROP_SOURCE] =
        g_param_spec_string ("source",
                             "Source",
                             "",
                             NULL,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpResource:target:
	 *
	 *
	 */
    properties[PROP_TARGET] =
        g_param_spec_string ("target",
                             "Target ",
                             "",
                             NULL,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpResource:overwrite:
	 *
	 *
     *
     * Recommended: %FALSE
	 */
    properties[PROP_OVERWRITE] =
        g_param_spec_boolean ("overwrite",
                              "Overwrite",
                              "",
                              FALSE,
                              G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpResource:total-bytes:
	 *
	 *
	 */
    properties[PROP_TOTAL_BYTES] =
        g_param_spec_uint64 ("total-bytes",
                             "Total Size",
                             "",
                             0,
                             G_MAXUINT64,
                             0,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpResource:downloaded-bytes:
	 *
	 *
	 */
    properties[PROP_DOWNLOADED_BYTES] =
        g_param_spec_uint64 ("downloaded-bytes",
                             "Downloaded Bytes",
                             "",
                             0,
                             G_MAXUINT64,
                             0,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpResource::last-progress-time:
	 *
	 *
	 */
    properties[PROP_LAST_PROGRESS_TIME] =
        g_param_spec_uint64 ("last-progress-time",
                             "Last Progress Time",
                             "",
                             0,
                             G_MAXUINT64,
                             0,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READABLE);

	/**
	 * SlurpResource::download-speed:
	 *
	 *
	 */
    properties[PROP_DOWNLOAD_SPEED] =
        g_param_spec_uint64 ("download-speed",
                             "Download Speed",
                             "",
                             0,
                             G_MAXUINT64,
                             0,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READABLE);

	/**
	 * SlurpResource::estimated-time:
	 *
	 *
	 */
    properties[PROP_ESTIMATED_TIME] =
        g_param_spec_uint64 ("estimated-time",
                             "Estimated Time",
                             "",
                             0,
                             G_MAXUINT64,
                             0,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READABLE);

    g_object_class_install_properties (object_class,
                                       NUM_PROPERTIES,
                                       properties);

    /**
     * SlurpResource::progress:
     * @object: the object which received the signal
     * @total: The total size of the resoure, 0 if no total size is set
     * @downloaded: The number of bytes downloaded
     *
     * The ::progress signal is emitted when the download progress of the
     * resource is :updated.
     */
    signals[PROGRESS] =
        g_signal_new ("progress",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      4,
                      G_TYPE_INT64,
                      G_TYPE_INT64,
                      G_TYPE_INT64,
                      G_TYPE_INT64);
}

static void
slurp_resource_init (SlurpResource *self)
{
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
SlurpResource *
slurp_resource_new (void)
{
    return g_object_new (SLURP_TYPE_RESOURCE, NULL);
}
void
slurp_resource_set_source (SlurpResource *self,
                           const gchar   *uri)
{
    _resource_set_source (self, uri);

    g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_SOURCE]);
}

gchar *
slurp_resource_get_source (SlurpResource *self)
{
    return _resource_get_source (self);
}

void
slurp_resource_set_target (SlurpResource *self,
                           const gchar   *uri)
{
    _resource_set_target (self, uri);

    g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_TARGET]);
}

gchar *
slurp_resource_get_target (SlurpResource *self)
{
    return _resource_get_target (self);
}


gboolean
slurp_resource_get_overwrite (SlurpResource *self)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    return priv->overwrite;
}

void
slurp_resource_set_overwrite (SlurpResource *self,
                              gboolean       overwrite)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    priv->overwrite = overwrite;

    g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_OVERWRITE]);
}

void
slurp_resource_set_total_bytes (SlurpResource *self,
                                gint64         bytes)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    priv->total_bytes = bytes;

    g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_TOTAL_BYTES]);
}

gint64
slurp_resource_get_total_bytes (SlurpResource *self)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    return priv->total_bytes;
}

void
slurp_resource_set_downloaded_bytes (SlurpResource *self,
                                     gint64         bytes)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    priv->downloaded_bytes = bytes;

    g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_DOWNLOADED_BYTES]);
}

void
slurp_resource_add_downloaded_bytes (SlurpResource *self,
                                     gint64         bytes)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    priv->downloaded_bytes += bytes;

    g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_DOWNLOADED_BYTES]);
}

gint64
slurp_resource_get_downloaded_bytes (SlurpResource *self)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    return priv->downloaded_bytes;
}

gint64
slurp_resource_get_download_speed (SlurpResource *self)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    return priv->download_speed;
}

gint64
slurp_resource_get_estimated_time (SlurpResource *self)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    return priv->estimated_time;
}

void
slurp_resource_update_progress (SlurpResource *self)
{
    _resource_calculate_progress (self, FALSE);
}

void
slurp_resource_force_update_progress (SlurpResource *self)
{
    _resource_calculate_progress (self, TRUE);
}
