/* slurp-resource.c
 *
 * Copyright © 2017 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "slurp-utils.h"

#include <string.h>

/**
 * slurp_utils_uri_is_directory:
 * @uri: a URI
 *
 * Checks if a supplied @uri is a directory.
 *
 * Returns: %TRUE if uri is a directory, %FALSE if not
 */
gboolean
slurp_utils_uri_is_directory (const gchar *uri)
{
    return g_str_has_suffix (uri, G_DIR_SEPARATOR_S);
}

/**
 * slurp_utils_create_directory:
 * @uri: a URI
 *
 * Creates the directory path of @uri, including the parent directories.
 *
 * Returns: %TRUE the directory path for @uri was created, %FALSE if not
 */
gboolean
slurp_utils_create_directory (const gchar *uri)
{
    gint retval;

    // TOOD: Add GError and get errno value from g_mkdir_with_parents()
    if (slurp_utils_uri_is_directory (uri))
        retval = g_mkdir_with_parents (uri, 0700);
    else
        // TODO: More error checking, g_path_get_dirname() can return '.'
        retval = g_mkdir_with_parents (g_path_get_dirname (uri), 0700);

    if (retval == 0)
        return TRUE;

    return FALSE;
}



char *
slurp_utils_uri_get_file_extention (const gchar *uri)
{
	gchar *p;

	if (uri == NULL)
		return NULL;

	p = g_strrstr (uri, ".");

	if (p == NULL)
		return NULL;

	if (p != uri)
    {
		gchar *p2;

		p2 = p - 1;
		while ((*p2 != '.') && (p2 != uri))
			p2--;

		if (strncmp (p2, ".tar.", 5) == 0)
			p = p2;
	}

	return g_strndup (p + 1, strlen (p) - 1);
}

char *
slurp_utils_uri_get_file_name (const gchar *uri)
{
    gchar *file_name = NULL;
	gchar *extension = NULL;

    file_name = g_strdup (g_path_get_basename (uri));

	if (uri == NULL)
		return NULL;

	extension = slurp_utils_uri_get_file_extention (uri);

	if (extension == NULL)
		return g_strdup (file_name);
	else
		return g_strndup (file_name, strlen (file_name) - (strlen (extension) + 1));
}

gchar *
slurp_utils_create_valid_target (const gchar *directory,
                                 const gchar *file_name,
                                 const gchar *extension,
                                 gboolean     overwrite,
                                 gint         increment)
{
    gchar *uri = NULL;

    if (increment > 0)
    {
        uri = g_build_filename (directory,
                                g_strconcat (file_name, "(", increment, ")", ".", extension, NULL),
                                NULL);
    }
    else
    {
        uri = g_build_filename (directory,
                                g_strconcat (file_name, ".", extension, NULL),
                                NULL);
    }

    g_print ("1");
    g_print ("\n\"%s\"\n", uri);
    g_print ("%s\n", g_file_test (uri, G_FILE_TEST_EXISTS) ? "true" : "false");
    if (g_file_test (uri, G_FILE_TEST_EXISTS))
    {
        g_print ("2");
        if (overwrite)
        {
            g_print ("4");
            return uri;
        }

        g_free (uri);
        g_print ("5");
        return slurp_utils_create_valid_target (directory, file_name, extension, overwrite, increment + 1);
    }
    else
    {
        g_print ("3");
        return uri;
    }
}

gchar *
slurp_utils_uri_create_valid_target (const gchar *uri,
                                     gboolean     overwrite)
{
    //TODO check for .file names
    gchar *directory = g_path_get_dirname (uri);
    gchar *file_name = slurp_utils_uri_get_file_name (uri);
    gchar *extension = slurp_utils_uri_get_file_extention (uri); //g_strrstr (filename, "."); // g_utf8_strrchr (filename_a, -1, '.');

    return slurp_utils_create_valid_target (directory, file_name, extension, overwrite, 0);
}