/* slurp-download.c
 *
 * Copyright © 2017 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define G_LOG_DOMAIN "slurp-download"

#include "slurp-download.h"

/**
 * SECTION:slurp-download
 * @title: SlurpDownload
 * @short_description: Asynchronous resource downloader using the libsoup library
 * @include: slurp-download.h
 * @see_also: #SoupSession, #SoupRequest, #GCallback, #GCancellable, #GFile,
 *            #GInputStream
 *
 * Asynchronous resource downloader using the libsoup library which offers
 * addded features over using plain `soup_*` calls.
 **/

typedef struct
{
    gchar *default_directory;
    gboolean default_overwrite;
    gchar *user_agent;

    SoupSession *session;
} SlurpDownloadPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpDownload, slurp_download, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_SESSION,
    PROP_USER_AGENT,
    PROP_DEFAULT_DIRECTORY,
    PROP_DEFAULT_OVERWRITE,
    NUM_PROPERTIES
};

static GParamSpec *properties[NUM_PROPERTIES];

GQuark
slurp_download_error_quark (void)
{
	static GQuark quark = 0;

	if (G_UNLIKELY (quark == 0))
	{
		quark = g_quark_from_static_string ("slurp-download-error");
	}

	return quark;
}

void
slurp_download_resource_splice_streams_close_async_cb (GObject      *object,
                                                       GAsyncResult *result,
                                                       gpointer      user_data)
{
    g_autoptr (GInputStream) input = NULL;
    g_autoptr (GError) error = NULL;
    gboolean retval = FALSE;
    GTask *task = NULL;
    SlurpDownloadData *data = NULL;

    g_return_if_fail (G_IS_INPUT_STREAM (object));
    g_return_if_fail (G_IS_TASK (user_data));

    input = G_INPUT_STREAM (object);
    task = G_TASK (user_data);
    data = g_task_get_task_data (task);

    retval = g_input_stream_close_finish (input, result, &error);

    slurp_resource_force_update_progress (data->resource);

    if (!retval)
    {
        g_debug ("Failed to close GInputStream");

        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_INPUT_STREAM_CLOSE_FAILED,
                                 "Failed to close GInputStream: %s",
                                 error->message);
        return;
    }

    g_debug ("Closing GInputStream");

    g_task_return_boolean (task, TRUE);
}

void
slurp_download_resource_splice_streams_async_read_cb (GObject      *object,
                                                      GAsyncResult *result,
                                                      gpointer      user_data)
{
    GInputStream *input = NULL; // Remove
    g_autoptr (GTask) task = NULL;
    g_autoptr (GError) error = NULL;
    GCancellable *cancellable = NULL;
    SlurpDownloadReadData *data = NULL;
    gsize nread;

    g_return_if_fail (G_IS_INPUT_STREAM (object));
    g_return_if_fail (G_IS_ASYNC_RESULT (result));
    g_return_if_fail (G_IS_TASK (user_data));
    g_return_if_fail (g_task_get_source_tag (G_TASK (user_data)) == slurp_download_resource_splice_streams_async);

    input = G_INPUT_STREAM (object);
    task = G_TASK (user_data);
    cancellable = g_task_get_cancellable (task);
    data = g_task_get_task_data (task);

    nread = g_input_stream_read_finish (G_INPUT_STREAM (object), result, &error);

    g_debug ("Reading new chunk from GInputStream");

    slurp_resource_update_progress (data->resource);

    if (nread == -1 || nread == 0 || error)
    {
        g_debug ("End of GInputStream");

        if (error)
        {
            g_task_return_new_error (task,
                                     SLURP_DOWNLOAD_ERROR,
                                     SLURP_DOWNLOAD_ERROR_INPUT_STREAM_NULL,
                                     "No GInputStream: %s",
                                     error->message);
        }
/*
        slurp_download_resource_splice_streams_close_async (self,
                                                            data->resource,
                                                            data->input,
                                                            data->output,
                                                            cancellable,
                                                            slurp_download_resource_splice_streams_close_async_cb,
                                                            g_object_ref (task));
*/


        g_debug ("Closing GInputStream");
        g_input_stream_close_async (input,
                                    G_PRIORITY_DEFAULT,
                                    g_task_get_cancellable (task),
                                    slurp_download_resource_splice_streams_close_async_cb,
                                    g_object_ref (task));

        // TODO: Add
        /*
        g_output_stream_close_async (g_object_ref (data->output),
                                     G_PRIORITY_DEFAULT,
                                     g_task_get_cancellable (task),
                                     slurp_download_resource_splice_streams_close_async_cb,
                                     g_object_ref (task));
        */
        return;
    }

    if (data->output)
    {
        gboolean written;
        gsize n_written;

        // TODO: Make async?
        written = g_output_stream_write_all (G_OUTPUT_STREAM (data->output),
                                             data->buffer,
                                             nread,
                                             &n_written,
                                             cancellable,
                                             &error);

        g_debug ("%s \"%lu\" bytes",
                 written ? "Successfully wrote" : "Failed to write",
                 n_written);

        if (error)
        {
            g_task_return_new_error (task,
                                     SLURP_DOWNLOAD_ERROR,
                                     SLURP_DOWNLOAD_ERROR_OUTPUT_STREAM_WRITE_FAILED,
                                     "Failed to write to target: %s",
                                     error->message);
        }

        slurp_resource_add_downloaded_bytes (data->resource, n_written);

        if (slurp_resource_get_downloaded_bytes (data->resource) > slurp_resource_get_total_bytes (data->resource))
        {
            slurp_resource_set_total_bytes (data->resource,
                                            slurp_resource_get_downloaded_bytes (data->resource));
        }

        if (!written)
        {
            g_input_stream_close_async (input,
                                        G_PRIORITY_DEFAULT,
                                        g_task_get_cancellable (task),
                                        slurp_download_resource_splice_streams_close_async_cb,
                                        g_object_ref (task));

            // TODO: Add g_output_stream_close_async ()

            return;
        }
    }
    else
    {
        g_debug ("GOutputStream no longer exists, quit");

        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_OUTPUT_STREAM_NULL,
                                 "No GOutputStream, can't write data");

        return;
    }

    g_input_stream_read_async (input,
                               data->buffer,
                               sizeof (data->buffer),
                               G_PRIORITY_DEFAULT,
                               g_task_get_cancellable (task),
                               slurp_download_resource_splice_streams_async_read_cb,
                               g_object_ref (task));
}

void
slurp_download_resource_splice_streams_async (SlurpDownload       *self,
                                              SlurpResource       *resource,
                                              GInputStream        *input,
                                              GOutputStream       *output,
                                              GCancellable        *cancellable,
                                              GAsyncReadyCallback  callback,
                                              gpointer             user_data)
{
    g_autoptr (GTask) task = NULL;
    SlurpDownloadReadData *data = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (SLURP_IS_RESOURCE (resource));
    g_return_if_fail (G_IS_INPUT_STREAM (input));
    g_return_if_fail (G_IS_OUTPUT_STREAM (output));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
    g_return_if_fail (G_IS_TASK (user_data));

    data = slurp_download_read_data_new ();
    data->resource = g_object_ref (resource);
    data->input = g_object_ref (input);
    data->output = g_object_ref (output);

    task = g_task_new (self, cancellable, callback, g_object_ref (user_data));
    g_task_set_source_tag (task, slurp_download_resource_splice_streams_async);
    g_task_set_task_data (task, data, g_object_unref);

    g_input_stream_read_async (data->input,
                               data->buffer,
                               sizeof (data->buffer),
                               G_PRIORITY_DEFAULT,
                               cancellable,
                               slurp_download_resource_splice_streams_async_read_cb,
                               g_object_ref (task));
}

gboolean
slurp_download_resource_splice_streams_finish (GObject       *object,
                                               GAsyncResult  *result,
                                               GError       **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (object), FALSE);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), FALSE);
    g_return_val_if_fail (g_task_is_valid (result, object), FALSE);
    g_return_val_if_fail (g_task_get_source_tag (G_TASK (result)) == slurp_download_resource_splice_streams_async, FALSE);

    return g_task_propagate_boolean (G_TASK (result), error);
}

void
slurp_download_resource_splice_streams_async_cb (GObject      *object,
                                                 GAsyncResult *result,
                                                 gpointer      user_data)
{
    g_autoptr (GError) error = NULL;
    gboolean retval = FALSE;
    GTask *task = NULL;
    //SlurpResource *resource = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (object));
    g_return_if_fail (G_IS_ASYNC_RESULT (result));
    g_return_if_fail (g_task_get_source_tag (G_TASK (result)) == slurp_download_resource_splice_streams_async);
    g_return_if_fail (G_IS_TASK (user_data));
    g_return_if_fail (g_task_get_source_tag (G_TASK (user_data)) == slurp_download_resource_async);

    task = G_TASK (user_data);
    //resource = g_task_get_task_data (task);

    retval = slurp_download_resource_splice_streams_finish (object, result, &error);

    if (!retval)
    {
        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_DOWNLOAD_FAILED,
                                 "Download failed: %s",
                                 error->message);

        return;
    }

    g_task_return_boolean (task, TRUE);
}

void
slurp_download_resource_create_input_async_send_cb (GObject      *object,
                                                    GAsyncResult *result,
                                                    gpointer      user_data)
{
    g_autoptr (SoupRequest) request = NULL;
    g_autoptr (GTask) task = NULL;
    g_autoptr (GInputStream) stream = NULL;
    g_autoptr (GError) error = NULL;
    SlurpResource *resource = NULL;

    g_return_if_fail (SOUP_IS_REQUEST (object));
    g_return_if_fail (G_IS_ASYNC_RESULT (result));
    g_return_if_fail (G_IS_TASK (user_data));
    g_return_if_fail (g_task_get_source_tag (G_TASK (user_data)) == slurp_download_resource_create_input_async);

    request = SOUP_REQUEST (object);
    task = G_TASK (user_data);
    resource = g_task_get_task_data (task);

    stream = soup_request_send_finish (request, result, &error);

    if (!stream)
    {
        g_task_return_error (task, error);
        return;
    }

    slurp_resource_set_total_bytes (g_object_ref (resource),
                                    (gint64) soup_request_get_content_length (g_object_ref (request)));

    g_task_return_pointer (task, g_object_ref (stream), g_object_unref);
}

void
slurp_download_resource_create_input_async (SlurpDownload       *self,
                                            SlurpResource       *resource,
                                            GCancellable        *cancellable,
                                            GAsyncReadyCallback  callback,
                                            gpointer             user_data)
{
    SlurpDownloadPrivate *priv = NULL;
    g_autoptr (GTask) task = NULL;
    g_autoptr (SoupSession) session = NULL;
    g_autoptr (SoupRequest) request = NULL;
    g_autoptr (GError) error = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
    g_return_if_fail (G_IS_TASK (user_data));
    g_return_if_fail (g_task_is_valid (user_data, self));

    priv = slurp_download_get_instance_private (self);

    task = g_task_new (self, cancellable, callback, g_object_ref (user_data));
    g_task_set_source_tag (task, slurp_download_resource_create_input_async);
    g_task_set_task_data (task, g_object_ref (resource), g_object_unref);

    if (!priv->session)
    {
        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_SOUP_SESSION_NULL,
                                 "No SoupSession set. Either failed to create SoupSession in SlurpDownload or NULL session passed.");

        return;
    }

    session = g_object_ref (priv->session);
    request = soup_session_request (session,
                                    slurp_resource_get_source (resource),
                                    &error);

    if (!request)
    {
        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_SOUP_REQUEST_NULL,
                                 "Failed to create SoupRequest for source URI");

        return;
    }

    soup_request_send_async (g_object_ref (request),
                             cancellable,
                             slurp_download_resource_create_input_async_send_cb,
                             g_object_ref (task));
}

GInputStream *
slurp_download_resource_create_input_finish (GObject       *object,
                                             GAsyncResult  *result,
                                             GError       **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (object), NULL);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), NULL);
    g_return_val_if_fail (g_task_is_valid (result, object), NULL);
    g_return_val_if_fail (g_task_get_source_tag (G_TASK (result)) == slurp_download_resource_create_input_async, NULL);

    return g_task_propagate_pointer (G_TASK (result), error);
}

void
slurp_download_resource_create_input_async_cb (GObject      *object,
                                               GAsyncResult *result,
                                               gpointer      user_data)
{
    g_autoptr (GError) error = NULL;
    g_autoptr (GTask) task = NULL;
    g_autoptr (GInputStream) input = NULL;
    SlurpDownloadData *data = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (object));

    g_return_if_fail (G_IS_TASK (result));
    g_return_if_fail (g_task_is_valid (result, object));
    g_return_if_fail (g_task_get_source_tag (G_TASK (result)) == slurp_download_resource_create_input_async);

    g_return_if_fail (G_IS_TASK (user_data));
    g_return_if_fail (g_task_is_valid (user_data, object));
    g_return_if_fail (g_task_get_source_tag (G_TASK (user_data)) == slurp_download_resource_async);

    task = G_TASK (user_data);
    data = g_task_get_task_data (task);
    input = slurp_download_resource_create_input_finish (object, result, &error);

    if (!input)
    {
        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_INPUT_STREAM_NULL,
                                 "Failed to create GInputStream for source URI: %s",
                                 error->message);
        return;
    }

    g_mutex_lock (&data->lock);
    data->input = g_object_ref (input);
    g_mutex_unlock (&data->lock);
}

void
slurp_download_resource_create_output_async (SlurpDownload       *self,
                                             SlurpResource       *resource,
                                             GCancellable        *cancellable,
                                             GAsyncReadyCallback  callback,
                                             gpointer             user_data)
{
    g_autoptr (GFile) file = NULL;
    g_autoptr (GTask) task = NULL;
    g_autoptr (GFileIOStream) iostream = NULL;
    g_autoptr (GOutputStream) stream = NULL;
    g_autoptr (GError) error = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (SLURP_IS_RESOURCE (resource));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
    g_return_if_fail (G_IS_TASK (user_data));
    g_return_if_fail (g_task_is_valid (user_data, self));

    task = g_task_new (self, cancellable, callback, g_object_ref (user_data));
    g_task_set_source_tag (task, slurp_download_resource_create_output_async);
    g_task_set_task_data (task, g_object_ref (resource), g_object_unref);

    // TODO: Ask sub GTask

    //g_file_create_readwrite

    file = g_file_new_for_uri (slurp_resource_get_target (resource));

    // TODO: Move to _async* versions
    if (slurp_resource_get_overwrite (resource))
    {
        iostream = g_file_replace_readwrite (file,
                                             NULL,
                                             FALSE,
                                             G_FILE_CREATE_REPLACE_DESTINATION,
                                             cancellable,
                                             &error);
    }
    else
    {
        iostream = g_file_create_readwrite (file,
                                            G_FILE_CREATE_NONE,
                                            cancellable,
                                            &error);
    }

    if (!iostream)
    {
        g_task_return_error (task, error);
        return;
    }

    /*
     * g_io_stream_get_output_stream() MUST take a g_object_ref() on iostream
     * since we are using g_autoptr(). As stated from the manual:
     *
     * > The GIOStream object owns the input and the output streams, not the
     * > other way around, so keeping the substreams alive will not keep the
     * > GIOStream object alive.
     *
     * stream pointer will hold a ref on the iostream pointer untill the stream
     * is g_object_unref()-ed.
     *
     * g_autoptr() is used incase a g_task_return_new_error() is called above,
     * automatically cleaning-up the iostream.
     */
    stream = g_io_stream_get_output_stream (g_object_ref (iostream));

    g_task_return_pointer (task, g_object_ref (stream), g_object_unref);
}

GOutputStream *
slurp_download_resource_create_output_finish (GObject       *object,
                                              GAsyncResult  *result,
                                              GError       **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (object), NULL);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), NULL);
    g_return_val_if_fail (g_task_is_valid (result, object), NULL);
    g_return_val_if_fail (g_task_get_source_tag (G_TASK (result)) == slurp_download_resource_create_output_async, NULL);

    return g_task_propagate_pointer (G_TASK (result), error);
}

void
slurp_download_resource_create_output_async_cb (GObject      *object,
                                                GAsyncResult *result,
                                                gpointer      user_data)
{
    g_autoptr (GTask) task = NULL;
    g_autoptr (GOutputStream) output = NULL;
    g_autoptr (GError) error = NULL;
    SlurpDownloadData *data = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (object));

    g_return_if_fail (G_IS_TASK (result));
    g_return_if_fail (g_task_is_valid (result, object));
    g_return_if_fail (g_task_get_source_tag (G_TASK (result)) == slurp_download_resource_create_output_async);

    g_return_if_fail (G_IS_TASK (user_data));
    g_return_if_fail (g_task_is_valid (user_data, object));
    g_return_if_fail (g_task_get_source_tag (G_TASK (user_data)) == slurp_download_resource_async);

    task = G_TASK (user_data);
    data = g_task_get_task_data (task);
    output = slurp_download_resource_create_output_finish (object, result, &error);

    // TODO: Will fail if path to file does not exists, create parent directories
    if (!output || output == NULL)
    {
        g_task_return_new_error (task,
                                 SLURP_DOWNLOAD_ERROR,
                                 SLURP_DOWNLOAD_ERROR_OUTPUT_STREAM_NULL,
                                 "Failed to create GOutputStream for target URI: %s",
                                 error->message);

        return;
    }

    g_mutex_lock (&data->lock);
    data->output = g_object_ref (output);
    g_mutex_unlock (&data->lock);
}

gboolean
slurp_download_resource_splice_streams_ready (GTask *task)
{
    SlurpDownloadData *data = NULL;
    gboolean retval = TRUE;

    g_return_val_if_fail (G_IS_TASK (task), FALSE);

    data = g_task_get_task_data (task);

    g_mutex_lock (&data->lock);

    if (!G_IS_INPUT_STREAM (data->input) || data->input == NULL)
    {
        retval = FALSE;
        goto out;
    }

    if (!G_IS_OUTPUT_STREAM (data->output) || data->output == NULL)
    {
        retval = FALSE;
        goto out;
    }

out:
    g_mutex_unlock (&data->lock);
    return retval;
}

typedef struct _SlurpDownloadResourceSpliceSource
{
    GSource         parent;
    GTask          *task;
    GDestroyNotify  destroy_message;
} SlurpDownloadResourceSpliceSource;

static gboolean
slurp_download_resource_splice_prepare (GSource *source,
                                        gint    *timeout)
{
    SlurpDownloadResourceSpliceSource *resource_slice_source = (SlurpDownloadResourceSpliceSource *) source;

    return slurp_download_resource_splice_streams_ready (resource_slice_source->task);
}

static gboolean
slurp_download_resource_splice_dispatch (GSource     *source,
                                         GSourceFunc  callback,
                                         gpointer     user_data)
{
    g_autoptr (GTask) task = NULL;
    SlurpDownload *self = NULL;
    GCancellable *cancellable = NULL;
    SlurpDownloadData *data = NULL;

    g_return_val_if_fail (G_IS_TASK (user_data), FALSE);

    task = G_TASK (user_data);
    self = g_task_get_source_object (task);
    cancellable = g_task_get_cancellable (task);
    data = g_task_get_task_data (task);

    slurp_download_resource_splice_streams_async (self,
                                                  data->resource,
                                                  data->input,
                                                  data->output,
                                                  cancellable,
                                                  slurp_download_resource_splice_streams_async_cb,
                                                  g_object_ref (task));

    return FALSE;
}

static void
slurp_download_resource_splice_finalize (GSource *source)
{
    SlurpDownloadResourceSpliceSource *resource_slice_source = (SlurpDownloadResourceSpliceSource *) source;

    g_object_unref (resource_slice_source->task);
}

static GSourceFuncs
slurp_download_resource_splice_funcs =
{
    slurp_download_resource_splice_prepare,
    NULL, // *_check() function, we don't need to use poll()
    slurp_download_resource_splice_dispatch,
    slurp_download_resource_splice_finalize,
    NULL
};

GSource *
slurp_download_resource_splice_source_new (SlurpDownload *self,
                                           GTask         *task,
                                           GCancellable  *cancellable)
{
    SlurpDownloadResourceSpliceSource *resource_slice_source;
    GSource *source;

    g_return_val_if_fail (SLURP_IS_DOWNLOAD (self), NULL);
    g_return_val_if_fail (G_IS_TASK (task), NULL);
    g_return_val_if_fail (g_task_is_valid (task, self), NULL);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), NULL);

    source = g_source_new (&slurp_download_resource_splice_funcs, sizeof (SlurpDownloadResourceSpliceSource));

    resource_slice_source = (SlurpDownloadResourceSpliceSource *) source;
    resource_slice_source->task = g_object_ref (task);

    g_source_set_name (source, "slurp-download-splice");

    if (cancellable)
    {
        GSource *cancellable_source;

        cancellable_source = g_cancellable_source_new (cancellable);
        g_source_add_child_source (source, cancellable_source);
        g_source_set_dummy_callback (cancellable_source);
        g_source_unref (cancellable_source);
    }

    return source;
}




// -----------------------------------------------------------------------------




void
slurp_download_resource_create_streams_async (SlurpDownload       *self,
                                              SlurpResource       *resource,
                                              GCancellable        *cancellable,
                                              GAsyncReadyCallback  callback,
                                              gpointer             user_data)
{
    g_autoptr (GTask) task = NULL;
    SlurpDownloadStreamData *data = NULL;

    data = slurp_download_stream_data_new ();
    data->resource = g_object_ref (resource);

    task = g_task_new (self, cancellable, callback, g_object_ref (user_data));
    g_task_set_source_tag (task, slurp_download_resource_create_streams_async);
    g_task_set_task_data (task, data, (GDestroyNotify) g_object_unref);

    slurp_download_resource_create_input_async (self,
                                                resource,
                                                cancellable,
                                                slurp_download_resource_create_input_async_cb,
                                                g_object_ref (task));

    slurp_download_resource_create_output_async (self,
                                                 resource,
                                                 cancellable,
                                                 slurp_download_resource_create_output_async_cb,
                                                 g_object_ref (task));

    // if (stream_ready)
    // {
    //     g_task_return_pointer (task, data, g_object_unref);
    // }
    // else
    // {
    //     add GSource
    // }


    GSource *source = NULL;

    source = slurp_download_resource_splice_source_new (self,
                                                        g_object_ref (task),
                                                        cancellable);

    g_task_attach_source (task,
                          source,
                          NULL);



    g_task_return_pointer (task, data, g_object_unref);
}

gpointer
slurp_download_resource_create_streams_finish (GObject       *object,
                                               GAsyncResult  *result,
                                               GError       **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (object), FALSE);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), FALSE);
    g_return_val_if_fail (g_task_is_valid (result, object), FALSE);
    g_return_val_if_fail (g_task_get_source_tag (G_TASK (result)) == slurp_download_resource_create_streams_async, FALSE);

    return g_task_propagate_pointer (G_TASK (result), error);
}


void
slurp_download_resource_splice_streams_ready_cb (GObject      *object,
                                                 GAsyncResult *result,
                                                 gpointer      user_data)
{
    g_autoptr (GTask) task = NULL;
    g_autoptr (GError) error = NULL;
    g_autoptr (SlurpDownloadStreamData) data = NULL;
    SlurpDownload *self = NULL;
    SlurpResource *resource = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (object));

    g_return_if_fail (G_IS_TASK (result));
    g_return_if_fail (g_task_is_valid (result, object));
    g_return_if_fail (g_task_get_source_tag (G_TASK (result)) == slurp_download_resource_create_streams_async);

    g_return_if_fail (G_IS_TASK (user_data));
    g_return_if_fail (g_task_is_valid (user_data, object));
    g_return_if_fail (g_task_get_source_tag (G_TASK (user_data)) == slurp_download_resource_async);

    task = G_TASK (user_data);
    self = g_task_get_source_object (task);
    resource = g_task_get_task_data (task);
    data = slurp_download_resource_create_streams_finish (object, result, &error);

    if (error)
    {
        g_task_return_error (task, error);

        return;
    }

    slurp_download_resource_splice_streams_async (self,
                                                  resource,
                                                  g_object_ref (data->input),
                                                  g_object_ref (data->output),
                                                  g_task_get_cancellable (task),
                                                  slurp_download_resource_splice_streams_async_cb,
                                                  g_object_ref (task));
}

void
slurp_download_resource_async2 (SlurpDownload       *self,
                               SlurpResource       *resource,
                               GCancellable        *cancellable,
                               GAsyncReadyCallback  callback,
                               gpointer             user_data)
{
    g_autoptr (GTask) task = NULL;
    SlurpDownloadData *data = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (SLURP_IS_RESOURCE (resource));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));

    data = slurp_download_data_new ();
    data->resource = resource;

    task = g_task_new (self, cancellable, callback, user_data);
    g_task_set_source_tag (task, slurp_download_resource_async);
    g_task_set_task_data (task, data, (GDestroyNotify) g_object_unref);

    slurp_download_resource_create_streams_async (self,
                                                  resource,
                                                  cancellable,
                                                  slurp_download_resource_splice_streams_ready_cb,
                                                  g_object_ref (task));
}





// -----------------------------------------------------------------------------








void
slurp_download_resource_async (SlurpDownload       *self,
                               SlurpResource       *resource,
                               GCancellable        *cancellable,
                               GAsyncReadyCallback  callback,
                               gpointer             user_data)
{
    g_autoptr (GTask) task = NULL;
    SlurpDownloadData *data = NULL;

    g_return_if_fail (SLURP_IS_DOWNLOAD (self));
    g_return_if_fail (SLURP_IS_RESOURCE (resource));
    g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));

    data = slurp_download_data_new ();
    data->resource = g_object_ref (resource);

    task = g_task_new (self, cancellable, callback, user_data);
    g_task_set_source_tag (task, slurp_download_resource_async);
    g_task_set_task_data (task, data, (GDestroyNotify) g_object_unref);

    // TODO: Create sub-task "task_create_stream" with GObject Boxed Type of SlurpDownloadStreamData as task_data.

    slurp_download_resource_create_input_async (self,
                                                resource,
                                                cancellable,
                                                slurp_download_resource_create_input_async_cb,
                                                g_object_ref (task));

    slurp_download_resource_create_output_async (self,
                                                 resource,
                                                 cancellable,
                                                 slurp_download_resource_create_output_async_cb,
                                                 g_object_ref (task));

    if (slurp_download_resource_splice_streams_ready (task))
    {
        slurp_download_resource_splice_streams_async (self,
                                                      data->resource,
                                                      data->input,
                                                      data->output,
                                                      cancellable,
                                                      slurp_download_resource_splice_streams_async_cb,
                                                      g_object_ref (task));
    }
    else
    {
        GSource *source = NULL;

        source = slurp_download_resource_splice_source_new (self,
                                                            g_object_ref (task),
                                                            cancellable);

        g_task_attach_source (task,
                              source,
                              NULL);
    }
}

gboolean
slurp_download_resource_finish (GObject       *object,
                                GAsyncResult  *result,
                                GError       **error)
{
    g_return_val_if_fail (SLURP_IS_DOWNLOAD (object), FALSE);
    g_return_val_if_fail (G_IS_ASYNC_RESULT (result), FALSE);
    g_return_val_if_fail (g_task_is_valid (result, object), FALSE);
    g_return_val_if_fail (g_task_get_source_tag (G_TASK (result)) == slurp_download_resource_async, FALSE);

    return g_task_propagate_boolean (G_TASK (result), error);
}

static void
slurp_download_finalize (GObject *object)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    g_clear_pointer (&priv->default_directory, g_free);
    g_clear_pointer (&priv->user_agent, g_free);
    g_clear_pointer (&priv->session, g_object_unref);

    G_OBJECT_CLASS (slurp_download_parent_class)->finalize (object);
}

static void
slurp_download_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SESSION:
            g_value_set_object (value, priv->session);
            break;

        case PROP_USER_AGENT:
            g_value_set_string (value, priv->user_agent);
            break;

        case PROP_DEFAULT_DIRECTORY:
            g_value_set_string (value, priv->default_directory);
            break;

        case PROP_DEFAULT_OVERWRITE:
            g_value_set_boolean (value, priv->default_overwrite);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    SlurpDownload *self = SLURP_DOWNLOAD (object);
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SESSION:
            priv->session = g_value_dup_object (value);
            break;

        case PROP_USER_AGENT:
            priv->user_agent = g_value_dup_string (value);
            break;

        case PROP_DEFAULT_DIRECTORY:
            priv->default_directory = g_value_dup_string (value);
            break;

        case PROP_DEFAULT_OVERWRITE:
            priv->default_overwrite = g_value_get_boolean (value);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_download_class_init (SlurpDownloadClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_download_finalize;
    object_class->get_property = slurp_download_get_property;
    object_class->set_property = slurp_download_set_property;

	/**
	 * SlurpDownload:session:
	 *
	 * A SoupSession which Slurp can use to download resources.
     *
     * Note: Make sure to ref your #SoupSession before passing it.
     *
     * g_object_new():
     * |[
     * g_object_new (SLURP_TYPE_DOWNLOAD,
     *               "session", g_object_ref (session),
     *               NULL);
     * ]|
     *
     * slurp_download_new_with_session():
     * |[
     * SlurpDownload *download;
     * download = slurp_download_new_with_session (g_object_ref (session));
     * ]|
	 */
    properties[PROP_SESSION] =
        g_param_spec_object ("session",
                             "A SoupSession",
                             "A SoupSession which Slurp can use to download resources.",
                             SOUP_TYPE_SESSION,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpDownload:user-agent:
	 *
	 * The user agent to pass when establishing a connection to the resource.
     *
     * Default: libslurp
	 */
    // TODO: Add new user-agent if session is passed
    properties[PROP_USER_AGENT] =
        g_param_spec_string ("user-agent",
                              "User Agent",
                              "The user agent to pass when establishing a connection to the resource.",
                              "slurp-download",
                              G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpDownload:directory:
	 *
	 * The default directory to save resources into.
     *
     * Default: $XDG_USER_DIR/DOWNLOAD (ex. $HOME/Downloads)
	 */
    properties[PROP_DEFAULT_DIRECTORY] =
        g_param_spec_string ("directory",
                             "Default directory",
                             "The default directory to save resources into.",
                             g_get_user_special_dir (G_USER_DIRECTORY_DOWNLOAD),
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);

	/**
	 * SlurpDownload:overwrite:
	 *
	 * Should files be overwritten by default.
     *
     * Recommended: %FALSE
	 */
    properties[PROP_DEFAULT_OVERWRITE] =
        g_param_spec_boolean ("overwrite",
                              "Overwrite files",
                              "Should files be overwritten by default? Default: FALSE",
                              FALSE,
                              G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);

    g_object_class_install_properties (object_class,
                                       NUM_PROPERTIES,
                                       properties);

}

static void
slurp_download_init (SlurpDownload *self)
{
    SlurpDownloadPrivate *priv = slurp_download_get_instance_private (self);

    if (!priv->session)
    {
        priv->session =
            soup_session_new_with_options (SOUP_SESSION_USER_AGENT, priv->user_agent,
                                           SOUP_SESSION_SSL_USE_SYSTEM_CA_FILE, TRUE,
                                           SOUP_SESSION_USE_THREAD_CONTEXT, TRUE,
                                           SOUP_SESSION_TIMEOUT, 60,
                                           SOUP_SESSION_IDLE_TIMEOUT, 60,
                                           NULL);
    }
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
SlurpDownload *
slurp_download_new (void)
{
    return g_object_new (SLURP_TYPE_DOWNLOAD, NULL);
}

SlurpDownload *
slurp_download_new_with_session (SoupSession *session)
{
    return g_object_new (SLURP_TYPE_DOWNLOAD,
                         "session", session,
                         NULL);
}
